#version 430 core

//https://learnopengl.com/PBR/Theory
//https://learnopengl.com/Advanced-Lighting/Normal-Mapping

in VertexData {
    vec3 position_world;
    vec3 normal_world;
    vec2 uv;
} vert;

uniform vec3 cameraWorldPosition;
//maps:
uniform sampler2D albedoMap;
uniform sampler2D normalMap;
uniform sampler2D metallicMap;
uniform sampler2D roughnessMap;
uniform sampler2D aoMap;

//brightness of material
uniform float brightness;

//lights:
uniform vec3 lightPositions[53];
uniform vec3 lightColors[53];


//output Color:
out vec4 FragColor;

const float PI = 3.14159265359;
//n
float DistributionGGX(vec3 N, vec3 H, float alpha)
{
    float nominator = alpha*alpha;
    float preSquare = (max(dot(N, H), 0.0)*max(dot(N, H), 0.0)*(nominator -1.0)+1.0);
    float denominator = PI*preSquare*preSquare;
    return nominator / denominator;
}
//n
float GeometrySchlickGGX(float NdotV, float alpha)
{
    float nom   = NdotV;
    float denom = NdotV * (1.0 - alpha) + alpha;
    return nom / denom;
}
//n
float GeometrySmith(vec3 N, vec3 V, vec3 L, float alpha)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx1 = GeometrySchlickGGX(NdotV, alpha);
    float ggx2 = GeometrySchlickGGX(NdotL, alpha);
	
    return ggx1 * ggx2;
}

//n
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

vec3 getNormalFromMap()
{
    vec3 tangentNormal = texture(normalMap, vert.uv).xyz * 2.0 - 1.0; //map to interval [-1.0 to 1.0]
    vec3 posDeriX  = dFdx(vert.position_world);//derivative of vertex x position
    vec3 posDeriY  = dFdy(vert.position_world);//derivative of vertex y position
    vec2 mapDeriX = dFdx(vert.uv);//derivative of uv in x
    vec2 mapDeriY = dFdy(vert.uv);//derivative of uv in y
    
    vec3 T  = normalize(posDeriX * mapDeriY.t - posDeriY * mapDeriX.t);
    vec3 N   = normalize(vert.normal_world);
    vec3 B  = -normalize(cross(T, N));
    mat3 TBN = mat3(T, B, N);

    return normalize(TBN * tangentNormal);
}

void main()
{		
    vec3  albedo    = pow(texture(albedoMap, vert.uv).rgb, vec3(2.2));
    float metallic  = texture(metallicMap, vert.uv).r;
    float roughness = texture(roughnessMap, vert.uv).r;
    float ao        = texture(aoMap, vert.uv).r;

    vec3 N = getNormalFromMap();
    vec3 V = normalize(cameraWorldPosition - vert.position_world);

	// calculate reflectance
    vec3 F0 = vec3(0.02); 
    F0 = mix(F0, albedo, metallic);
    vec3 Lo = vec3(0.0);
    for(int i = 0; i < lightPositions.length; ++i) 
    {
        //calculate the radiance for each light
        vec3 L = normalize(lightPositions[i] - vert.position_world);
        vec3 H = normalize(V + L);
        float distance = length(lightPositions[i] - vert.position_world);
        //calculate attenuation
        float attenuation = 1.0 / (distance * distance);
        //calculate radiance for each light
        vec3 radiance = lightColors[i] * attenuation;

        // Cook-Torrance Model
        float NDF = DistributionGGX(N, H, roughness);   
        float G = GeometrySmith(N, V, L, roughness);      
        vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);
        vec3 nominator    = NDF * G * F; 
        float denominator = 4 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.001;
        vec3 specular = nominator / denominator;
        
        vec3 kS = F;
        vec3 kD = (vec3(1.0) - kS) * (1.0 - metallic);	  
        float NdotL = max(dot(N, L), 0.0);         

        Lo += (kD * albedo / PI + specular) * radiance * NdotL; 
    }   
    
    vec3 ambient = vec3(0.03) * albedo * ao * brightness;
	// Final Color
    vec3 color = ambient + Lo;
    color = color / (color + vec3(1.0));
    color = pow(color, vec3(1.0/2.2)); 
    FragColor = vec4(color, 1.0);
}